import fcntl
import os

def _APPEND_PAGE(_DB, _PAGE_SIZE):
    _DB.write(b'0'*_PAGE_SIZE)
    _DB.flush()
    return


def _DELETE_PAGE(_DB, _PAGE_SIZE):
    _DB.seek(0, 2) # 移到文件末尾
    _LENGTH = _DB.tell()
    _DB.truncate(_LENGTH-_PAGE_SIZE) # 往前截断一个page的大小,如果文件原本大小不足page_size的话则会报错
    _DB.flush()
    return


def _PARSE_SINGLE_PAGE(_DB, _OFFSET, _PAGE_SIZE, _RET, _INDEX_LEN, _LOG_LEN): # 0:_INDEX 1:_LOG 2:_DATA
    """
    用于解析当前页面的内容，页面的框选通过跨页面偏移offset实现
    :param _DB:
    :param _OFFSET:
    :param _PAGE_SIZE:
    :param _RET:
    :param _INDEX_LEN:
    :param _LOG_LEN:
    :return:
    """
    if _RET == 0x00:
        _DB.seek(_OFFSET)
        _INDEX_SLICE = _DB.read(_INDEX_LEN)
        return _INDEX_SLICE
    elif _RET == 0x01:
        # print('seek here', _OFFSET+_INDEX_LEN)
        _DB.seek(_OFFSET+_INDEX_LEN)
        _LOG_SLICE = _DB.read(_LOG_LEN)
        return _LOG_SLICE
    elif _RET == 0x02:
        # print(_OFFSET+_INDEX_LEN+_LOG_LEN)
        _DB.seek(_OFFSET+_INDEX_LEN+_LOG_LEN)
        _CONTENT = _DB.read(_PAGE_SIZE-_INDEX_LEN-_LOG_LEN)
        return _CONTENT

def _CONTENT_MULTIPAGE(_DB, _CONTENT_LEN, _HEAD_SIZE=1024, _PAGE_SIZE=1024, _RET=0, _INDEX_LEN=0x100, _LOG_LEN=0x100, _START_OFFSET=None):
    """
    调用parse single page 实现对于某一内容在所有页面上的读取，需告知内容的总长度
    :param _DB:
    :param _CONTENT_LEN:
    :param _HEAD_SIZE:
    :param _PAGE_SIZE:
    :param _RET:
    :param _INDEX_LEN:
    :param _LOG_LEN:
    :return:
    """
    _CONTNET = b''
    _OVER = False
    _START_PAGE = 0
    if _RET == 0x00:
        _PAGE_NUMS = _CONTENT_LEN // _INDEX_LEN
        if _CONTENT_LEN % _INDEX_LEN != 0:
            _OVER = True

    elif _RET == 0x01:
        _PAGE_NUMS = _CONTENT_LEN // _LOG_LEN
        if _CONTENT_LEN % _LOG_LEN != 0:
            _OVER = True


    elif _RET == 0x02:
        _PAGE_NUMS = _CONTENT_LEN // (_PAGE_SIZE-_INDEX_LEN-_LOG_LEN)
        if _CONTENT_LEN % (_PAGE_SIZE-_INDEX_LEN-_LOG_LEN) != 0:
            _OVER = True
        if _START_OFFSET is None:
            raise Exception('StartOffestError', '_START_OFFSET is needed when requesting data field.')
        if _START_OFFSET > 0:
            _START_PAGE = _START_OFFSET // (_PAGE_SIZE-_INDEX_LEN-_LOG_LEN)
            _PAGE_INNER_OFFSET = _START_OFFSET % (_PAGE_SIZE-_INDEX_LEN-_LOG_LEN)
            _DB.seek(_HEAD_SIZE+_START_PAGE*_PAGE_SIZE+_LOG_LEN+_INDEX_LEN+_PAGE_INNER_OFFSET)
            _READ_LEN = min(_CONTENT_LEN, (_PAGE_SIZE-_LOG_LEN-_INDEX_LEN-_PAGE_INNER_OFFSET))
            _CONTNET += _DB.read(_READ_LEN)
            _START_PAGE += 1
            # print('prefix', _CONTNET)
            # quit()

    # print('pn', _PAGE_NUMS)
    # print('sp', _START_PAGE)

    if _OVER:
        _PAGE_NUMS += 1
    for _PAGE_NUM in range(_START_PAGE, _START_PAGE+_PAGE_NUMS, 0x01):
        _CURR_OFFSET = _HEAD_SIZE + _PAGE_NUM * _PAGE_SIZE
        # print(_CURR_OFFSET)
        _CONTENT_BUFFER = _PARSE_SINGLE_PAGE(_DB, _OFFSET=_CURR_OFFSET, _PAGE_SIZE=_PAGE_SIZE, _RET=_RET, _INDEX_LEN=_INDEX_LEN, _LOG_LEN=_LOG_LEN)
        # print(_CONTENT_BUFFER)
        _CONTNET += _CONTENT_BUFFER
    return _CONTNET[:_CONTENT_LEN]



def _DB_WRITE_BYTES(_DB, _OFFSET, _CONTENT, _PAGE_LIMIT):
    """
    直接向文件写入一串字节，通过page_limit进行页面溢出检查
    :param _DB:
    :param _OFFSET:
    :param _CONTENT:
    :param _PAGE_LIMIT:
    :return:
    """
    # print('file_offset', _OFFSET)
    if _OFFSET+len(_CONTENT) > _PAGE_LIMIT:
        raise Exception('PageOverflow')
    _DB.seek(_OFFSET)
    # fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
    _DB.write(_CONTENT)
    # fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
    _DB.flush()
    return _DB.tell()

def _GET_TOTAL_SIZE(_DB):
    _DB.seek(0, 2)
    return _DB.tell()

def _WRITE_IN_PAGE(_DB, _CONTENT_BUFFER, _PAGE_NUM,_PAGE_OFFSET, _PAGE_SIZE, _HEAD_SIZE,_INDEX_LEN, _LOG_LEN, _TYPE):
    """
    在指定页面写内容，做安全检查，最终通过db_write_content实现，在本函数中需要计算当前页面末尾位置以及contnet buffer的大小从而确认安全
    传入当前页面号，页面偏移便可知道
    该函数计算页面偏移和页内偏移并相加作为总偏移量
    :param _CONTNET_BUFFER:
    :param _PAGE_NUM:
    :param __PAGE_SIZE:
    :param _HEAD_SIZE:
    :param _INDEX_LEN:
    :param _LOG_LEN:
    :return:
    """

    if _TYPE == 0x00:
        if not _PAGE_OFFSET == 0x00:
            raise Exception('PageInnerOffsetError')
        if len(_CONTENT_BUFFER) > _INDEX_LEN:
            raise Exception('INDEXContentOverflow')
        _PAGE_NUM_OFFSET = _PAGE_NUM * _PAGE_SIZE + _HEAD_SIZE
        _CURR_SIZE = _GET_TOTAL_SIZE(_DB)
        _CURR_TOTAL_PAGE = (_CURR_SIZE - _HEAD_SIZE) // _PAGE_SIZE
        if _CURR_TOTAL_PAGE <= _PAGE_NUM:
            _APPEND_PAGE(_DB, _PAGE_SIZE=_PAGE_SIZE)
        return _DB_WRITE_BYTES(_DB, _PAGE_NUM_OFFSET, _CONTENT=_CONTENT_BUFFER, _PAGE_LIMIT=_PAGE_NUM_OFFSET+_PAGE_SIZE)

    if _TYPE == 0x01:
        if not _PAGE_OFFSET == _INDEX_LEN:
            raise Exception('PageInnerOffsetError')
        if len(_CONTENT_BUFFER) > _LOG_LEN:
            raise Exception('LOGContentOverflow')
        _CURR_SIZE = _GET_TOTAL_SIZE(_DB)
        _CURR_TOTAL_PAGE = (_CURR_SIZE - _HEAD_SIZE) // _PAGE_SIZE
        if _CURR_TOTAL_PAGE <= _PAGE_NUM:
            _APPEND_PAGE(_DB, _PAGE_SIZE=_PAGE_SIZE)
        _PAGE_NUM_OFFSET = _PAGE_NUM * _PAGE_SIZE + _HEAD_SIZE + _INDEX_LEN
        return _DB_WRITE_BYTES(_DB, _PAGE_NUM_OFFSET, _CONTENT=_CONTENT_BUFFER, _PAGE_LIMIT=_PAGE_NUM_OFFSET+_LOG_LEN)

    if _TYPE == 0x02:
        if not _PAGE_OFFSET == _INDEX_LEN + _LOG_LEN:
            raise Exception('PageInnerOffsetError')
        if len(_CONTENT_BUFFER) > _PAGE_SIZE - _LOG_LEN - _INDEX_LEN:
            raise Exception('DataContentOverflow')
        _CURR_SIZE = _GET_TOTAL_SIZE(_DB)
        _CURR_TOTAL_PAGE = (_CURR_SIZE - _HEAD_SIZE) // _PAGE_SIZE
        if _CURR_TOTAL_PAGE <= _PAGE_NUM:
            _APPEND_PAGE(_DB, _PAGE_SIZE=_PAGE_SIZE)
        _PAGE_NUM_OFFSET = _PAGE_NUM * _PAGE_SIZE + _HEAD_SIZE + _INDEX_LEN + _LOG_LEN
        # print('page_num_offset', _PAGE_NUM_OFFSET, _PAGE_NUM_OFFSET+len(_CONTENT_BUFFER))
        return _DB_WRITE_BYTES(_DB, _PAGE_NUM_OFFSET, _CONTENT=_CONTENT_BUFFER, _PAGE_LIMIT=_PAGE_NUM_OFFSET+(_PAGE_SIZE-_LOG_LEN-_INDEX_LEN))


def _LOCK_FILE(_DB):
    fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
    return


def _UNLOCK_FILE(_DB):
    fcntl.flock(_DB.fileno(), fcntl.LOCK_UN)
    return


def _CONTENT_SET(_CONTENT, _DB, _PAGE_SIZE=1024, _TYPE=0x00, _INDEX_LEN=0x100, _LOG_LEN=0x100, _HEAD_SIZE=1024, _LOG_OFFSET=None, _DATA_OFFSET=None):
    """
    将传入的字节流持久化到数据库中。可以持久化INDEX、LOG、DATA，对于较长字节流，支持自动判断、换页存储，循环写；并在循环外加互斥锁（调用write_in_page进而调用DBwritebytes实现写功能）

    LOG_OFFSET:原有的log长度
    :param _CONTENT:
    :param _DB:
    :param _PAGE_SIZE:
    :param _TYPE:
    :param _INDEX_LEN:
    :param _LOG_LEN:
    :return:
    """
    assert type(_CONTENT) == bytes
    _CONTENT_LEN = len(_CONTENT)
    # print(_CONTENT_LEN)
    if _TYPE == 0x00:
        _PAGE_NUM = 0x00
        # fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
        for _CONTENT_OFFSET in range(0, _CONTENT_LEN, _INDEX_LEN):
            _CONTENT_BUFFER = _CONTENT[_CONTENT_OFFSET:_CONTENT_OFFSET+_INDEX_LEN]
            # print(_PAGE_NUM, _CONTENT_BUFFER)
            _WRITE_IN_PAGE(_DB=_DB,
                           _CONTENT_BUFFER=_CONTENT_BUFFER,
                           _PAGE_NUM=_PAGE_NUM,
                           _PAGE_OFFSET=0x00,
                           _PAGE_SIZE=_PAGE_SIZE,
                           _INDEX_LEN=_INDEX_LEN,
                           _LOG_LEN=_LOG_LEN,
                           _HEAD_SIZE=_HEAD_SIZE,
                           _TYPE=_TYPE
                           )
            _PAGE_NUM += 0x01
        fcntl.flock(_DB.fileno(), fcntl.LOCK_UN)

    if _TYPE == 0x01:
        if _LOG_OFFSET is None:
            raise Exception('LOG_OFFSET is needed.')
        # 计算从第几页开始
        _START_PAGE = _LOG_OFFSET // _LOG_LEN
        _OVER_OFFSET = _LOG_OFFSET % _LOG_LEN
        _PAGE_NUM = _START_PAGE
        # fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
        _CURR_INDEX = None

        # print(_PAGE_NUM, _OVER_OFFSET, _START_PAGE)

        if _OVER_OFFSET > 0x00:
            _DB_WRITE_BYTES(_DB,
                            _OFFSET=_PAGE_NUM*_PAGE_SIZE+_OVER_OFFSET+_INDEX_LEN+_HEAD_SIZE,
                            _CONTENT=_CONTENT[:_LOG_LEN-_OVER_OFFSET],
                            _PAGE_LIMIT=_PAGE_NUM*_PAGE_SIZE+_LOG_LEN+_INDEX_LEN+_HEAD_SIZE)
            _PAGE_NUM += 0x01
            for _LOG_CONTENT_IDX in range(_LOG_LEN-_OVER_OFFSET, len(_CONTENT), _LOG_LEN):
                _CURR_INDEX = _WRITE_IN_PAGE(_DB,
                               _CONTENT_BUFFER=_CONTENT[_LOG_CONTENT_IDX:_LOG_CONTENT_IDX+_LOG_LEN],
                               _PAGE_NUM=_PAGE_NUM,
                               _PAGE_OFFSET=_INDEX_LEN,
                               _PAGE_SIZE=_PAGE_SIZE,
                               _HEAD_SIZE=_HEAD_SIZE,
                               _INDEX_LEN=_INDEX_LEN,
                               _LOG_LEN=_LOG_LEN,
                               _TYPE=0x01)
                _PAGE_NUM += 0x01
        else:
            for _LOG_CONTENT_IDX in range(0, len(_CONTENT), _LOG_LEN):
                _CURR_INDEX = _WRITE_IN_PAGE(_DB,
                               _CONTENT_BUFFER=_CONTENT[_LOG_CONTENT_IDX:_LOG_CONTENT_IDX+_LOG_LEN],
                               _PAGE_NUM=_PAGE_NUM,
                               _PAGE_OFFSET=_INDEX_LEN,
                               _PAGE_SIZE=_PAGE_SIZE,
                               _HEAD_SIZE=_HEAD_SIZE,
                               _INDEX_LEN=_INDEX_LEN,
                               _LOG_LEN=_LOG_LEN,
                               _TYPE=0x01)
                _PAGE_NUM += 0x01
        fcntl.flock(_DB.fileno(), fcntl.LOCK_UN)
        return _LOG_OFFSET+len(_CONTENT), _CURR_INDEX  # 当前log总产度，当前指针位置

    if _TYPE == 0x02:
        if _DATA_OFFSET is None:
            raise Exception('DATA_OFFSET is needed.')
        # 计算从第几页开始
        _START_PAGE = _DATA_OFFSET // (_PAGE_SIZE-_LOG_LEN-_INDEX_LEN)
        _OVER_OFFSET = _DATA_OFFSET % (_PAGE_SIZE-_LOG_LEN-_INDEX_LEN)
        _PAGE_NUM = _START_PAGE
        # fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
        _CURR_INDEX = None
        # print(_PAGE_NUM, _OVER_OFFSET, _START_PAGE)
        # print(_PAGE_NUM, _OVER_OFFSET, _START_PAGE)

        if _OVER_OFFSET > 0x00:
            _DB_WRITE_BYTES(_DB,
                            _OFFSET=_PAGE_NUM * _PAGE_SIZE + _OVER_OFFSET + _INDEX_LEN + _HEAD_SIZE + _LOG_LEN,
                            _CONTENT=_CONTENT[:(_PAGE_SIZE - _LOG_LEN - _INDEX_LEN) - _OVER_OFFSET],
                            _PAGE_LIMIT=(_PAGE_NUM+1) * _PAGE_SIZE + _HEAD_SIZE)
            _PAGE_NUM += 0x01
            for _DATA_CONTENT_IDX in range((_PAGE_SIZE - _LOG_LEN - _INDEX_LEN) - _OVER_OFFSET, len(_CONTENT), _PAGE_SIZE - _LOG_LEN - _INDEX_LEN):
                _CURR_INDEX = _WRITE_IN_PAGE(_DB,
                                             _CONTENT_BUFFER=_CONTENT[_DATA_CONTENT_IDX:_DATA_CONTENT_IDX + (_PAGE_SIZE - _LOG_LEN - _INDEX_LEN)],
                                             _PAGE_NUM=_PAGE_NUM,
                                             _PAGE_OFFSET=_INDEX_LEN+_LOG_LEN,
                                             _PAGE_SIZE=_PAGE_SIZE,
                                             _HEAD_SIZE=_HEAD_SIZE,
                                             _INDEX_LEN=_INDEX_LEN,
                                             _LOG_LEN=_LOG_LEN,
                                             _TYPE=0x02)
                _PAGE_NUM += 0x01
        else:
            for _DATA_CONTENT_IDX in range(0, len(_CONTENT), _PAGE_SIZE - _LOG_LEN - _INDEX_LEN):
                _CURR_INDEX = _WRITE_IN_PAGE(_DB,
                                             _CONTENT_BUFFER=_CONTENT[_DATA_CONTENT_IDX:_DATA_CONTENT_IDX + (_PAGE_SIZE - _LOG_LEN - _INDEX_LEN)],
                                             _PAGE_NUM=_PAGE_NUM,
                                             _PAGE_OFFSET=_INDEX_LEN+_LOG_LEN,
                                             _PAGE_SIZE=_PAGE_SIZE,
                                             _HEAD_SIZE=_HEAD_SIZE,
                                             _INDEX_LEN=_INDEX_LEN,
                                             _LOG_LEN=_LOG_LEN,
                                             _TYPE=0x02)
                _PAGE_NUM += 0x01
        # fcntl.flock(_DB.fileno(), fcntl.LOCK_UN)
        return _DATA_OFFSET + len(_CONTENT), _CURR_INDEX  # 当前log总产度，当前指针位置






    """
    TODO:
    计算页面号
    将传入的字节流切片并一块一块通过调用write_in_page存放入对应的页面号中,循环之前加锁，循环后释放
    """

def _WRITE_EMPTY_HEAD(_DB, _HEAD_SIZE):
    _DB.write(b'0'*_HEAD_SIZE)
    _DB.flush()
    return 0x00

def _CONN_DATABASE(_NAME, _HEAD_SIZE=None):
    if os.path.exists(_NAME):
        _DB = open(_NAME, 'r+b')
        return _DB, True

    else:
        if _HEAD_SIZE is None:
            raise Exception('DBInitError', 'DBHeadSizeNotAssign')
        _DB = open(_NAME, 'wb')
        fcntl.flock(_DB.fileno(), fcntl.LOCK_EX)
        _WRITE_EMPTY_HEAD(_DB, _HEAD_SIZE)
        fcntl.flock(_DB.fileno(), fcntl.LOCK_UN)
        _DB.close()
        _DB = open(_NAME, 'r+b')
        return _DB, False

def _CLOSE_DB(_DB):
    _DB.flush()
    _DB.close()
    return None

def _READ_HEAD(_DB, _HEAD_SIZE):
    _DB.seek(0)
    _HEAD_CONTENT = _DB.read(_HEAD_SIZE)
    return _HEAD_CONTENT

def _WRITE_HEAD(_DB, _HEAD_CONTENT, _HEAD_SIZE):
    _DB.seek(0)
    _LOCK_FILE(_DB)
    if len(_HEAD_CONTENT) >= _HEAD_SIZE:
        raise Exception('HeadContentOVerflow')
    _DB.write(b'0'*_HEAD_SIZE)
    _DB.seek(0)
    _DB.write(_HEAD_CONTENT)
    _DB.flush()
    _UNLOCK_FILE(_DB)
    pass


# HEAD STRUCTURE: 4 _HEAD-SIZE:DICT




if __name__ == '__main__':

   pass



from Odyssey.interface import Session
from Odyssey.db_conf import DBConf
import os
import time
import warnings
import configparser
import fcntl

class Handler:
    def __init__(self, dbname, load_config=None):
        self.dbname = dbname
        log_len, idx_len, head_size, page_size, version = self.load_config(config_path=load_config)
        if os.path.exists(dbname):
            if load_config is not None:
                warnings.warn('The db is exist. Ignored the input configs.')
            self.db, status = Session.connect(db_name=self.dbname, head_size=None)
        else:
            self.db, status = Session.connect(db_name=self.dbname, head_size=head_size)
        if status:
            self.session = Session(self.db)
            self.session.load_config_from_head()
        else:
            self.session = Session(self.db, _HEAD_SIZE=head_size, _PAGE_SIZE=page_size, _IDX_LEN=idx_len, _LOG_LEN=log_len)
            self.session.compile_head(idx_len=idx_len, log_len=log_len, head_size=head_size, page_size=page_size, version=version, info=None)
    def load_config(self, config_path=None):
        if not config_path:
            log_len = DBConf.log_len
            idx_len = DBConf.idx_len
            head_size = DBConf.head_size
            page_size = DBConf.page_size
            version = DBConf.version
        else:
            cfg = configparser.ConfigParser()
            cfg.read(config_path)
            # print(cfg_dic)
            log_len = int(cfg.get('DB', 'log_len'))
            idx_len = int(cfg.get('DB', 'idx_len'))
            head_size = int(cfg.get('DB', 'head_size'))
            page_size = int(cfg.get('DB', 'page_size'))
            version = cfg.get('DB', 'version')
            # quit()
        return log_len, idx_len, head_size, page_size, version

    def get(self, key):
        index = self.session.get_index()
        if key in index:
            data_idx = index[key]
            data_dic = self.session.get_data(data_idx)
            if data_dic['exp'] is None or data_dic['exp'] >= (time.time() - data_dic['set_time']):
                return data_dic['data']
            # raise Exception('KeyError', 'Key is out of date.')
        raise Exception('KeyError', 'Key Not In Database Or Out Of Date.')
        # return None

    def set(self, key, value, exp=None): # 多进程不安全
        self.lock_db()
        index = self.session.get_index()
        if key in index:
            last_idx = index[key]
            time_stamp = time.time()
            index[key] = self.session.set_data({'data':value, 'last_idx':last_idx, 'exp':exp, 'set_time':time_stamp}, key)
        else:
            time_stamp = time.time()
            index[key] = self.session.set_data({'data':value, 'last_idx':None, 'exp':exp, 'set_time':time_stamp}, key)
        self.session.update_index(index)
        self.unlock_db()

    def get_history(self, key, deep=None, back_time=None):
        index = self.session.get_index()
        hist_dic = {}
        count = 0
        if key in index:
            capsule = self.session.get_data(index[key])
            while capsule['last_idx'] is not None:
                count += 1
                if deep and count > deep:
                    break
                if back_time and back_time < time.time() - capsule['set_time']:
                    break
                hist_dic[capsule['set_time']] = capsule['data']
                capsule = self.session.get_data(capsule['last_idx'])
        return hist_dic

    def check_log(self, desc=True, limit=None):
        logs = self.session.check_log()
        return logs

    def set_head_info(self, info):
        self.lock_db()
        head = self.session.get_head()
        head['info'] =info
        self.session.set_head(head)
        self.unlock_db()

    def get_head(self):
        return self.session.get_head()

    def del_key(self, key):
        self.lock_db()
        index = self.session.get_index()
        if key in index:
            last_idx = index[key]
            time_stamp = time.time()
            index[key] = self.session.set_data(
                {'data': None, 'last_idx': last_idx, 'exp': 0, 'set_time': time_stamp}, key)
            self.session.update_index(index)
        else:
            pass
        self.unlock_db()

    def get_all_dic(self):
        index = self.session.get_index()
        dic = {}
        for k in index:
            try:
                dic[k] = self.get(k)
            except:
                pass
        return dic

    def lock_db(self):
        fcntl.flock(self.db.fileno(), fcntl.LOCK_EX)

    def unlock_db(self):
        fcntl.flock(self.db.fileno(), fcntl.LOCK_UN)

if __name__ == '__main__':

    h = Handler('test5.db')
    # h.set('im', 'lyx')
    # print(h.get('im'))
    # print(h.set('u', 12222222222222222222222222, exp=10))
    # print(h.session.get_index())
    # print(h.get('u'))
    print(h.get_history('u', back_time=100000000, deep=3)) # ok


import os
from Odyssey.handler import Handler
# from Odyssey.interface import Session

def connect(dbname, use_config=None):
    db = Odyssey(dbname, use_config)
    return db

class Odyssey(object):
    def __init__(self, db_name, use_config=None):
        self.db_path = db_name
        self.handler = Handler(db_name, load_config=use_config)

    def set(self, key, value, exp=None):
        self.handler.set(key, value, exp)

    def get(self, key):
        return self.handler.get(key)

    def get_history(self, key, deep=None, limit_time=None):
        return self.handler.get_history(key, deep=deep, back_time=limit_time)

    def get_log(self, desc=True, limit=None):
        return self.handler.check_log(desc=desc, limit=limit)

    def __setitem__(self, key, value):
        self.handler.set(key, value)

    def __getitem__(self, key):
        return self.handler.get(key)

    def __call__(self, info):
        self.handler.set_head_info(info)
        pass

    def __delitem__(self, key):
        self.del_key(key)

    def summery(self): # 换成.summery,打印就打印一个dict
        head_info = self.handler.get_head()
        disk_use = os.path.getsize(self.db_path)//1024
        print("""
    -*- Odyssey Database Version {} -*-
     |                               |
     |   ```                         |
     |       INFO:{}               |
     |                               |
     |       LOG_LEN:{}              |
     |       IDX_LEN:{}              |   
     |       HEAD_SIZE:{}          |
     |       PAGE_SIZE:{}            |
     |                               |
     |       Disk Occupy:{} KB        |   
     |   ```                         |
     ---------------------------------
        """.format(head_info['version'], head_info['info'], head_info['LOG_LEN'], head_info['IDX_LEN'], head_info['HEAD_SIZE'], head_info['PAGE_SIZE'], disk_use))
        return {'version':head_info['version'],
                'info':head_info['info'],
                'log_len':head_info['LOG_LEN'],
                'idx_len':head_info['IDX_LEN'],
                'head_size':head_info['HEAD_SIZE'],
                'page_size':head_info['PAGE_SIZE'],
                'disk_use': disk_use
                }

    def get_info(self):
        return self.handler.get_head()['info']

    def del_key(self, key):
        self.handler.del_key(key)

    def get_all(self):
        return self.handler.get_all_dic()

    def __str__(self):
        return '{}'.format(self.get_all())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handler.session.close()

    def __iter__(self):
        keys = self.handler.get_all_dic().keys()
        yield from keys

    def __len__(self):
        return len(self.handler.get_all_dic())

    def __sizeof__(self):
        return os.path.getsize(self.db_path)

    def __eq__(self, other):
        if self.handler.get_all_dic() == other.get_all():
            return True
        return False

    def __contains__(self, item):
        if item in self.handler.get_all_dic():
            return True

    def keys(self):
        return self.handler.get_all_dic().keys()

    def values(self):
        return self.handler.get_all_dic().values()

    def pop(self, key):
        last_val = self.get(key)
        self.del_key(key)
        return last_val

    def __bool__(self):
        return not self.handler.db.closed

    def close(self):
        self.handler.session.close()













from Odyssey.file_utils import _CONTENT_MULTIPAGE, _CONTENT_SET, _WRITE_HEAD, _READ_HEAD, _CONN_DATABASE
import struct
import pickle
import time

class Session:
    """
    Odyssey（提供redis-like api）
    |
    Handler（日志对其透明，但必要时可以访问；）
    |
    Session（封装底层文件处理函数；封装日志系统，日志自动记录；） Session
    |
    FILE BASEMENT

    Interface相当于打开一个session，是面向一个连接的接口，传入的是连接，可以调用底层文件系统进行接近底层的读写封装


    """
    def __init__(self, _DB, _HEAD_SIZE=None, _PAGE_SIZE=None, _IDX_LEN=None, _LOG_LEN=None):
        self.db = _DB
        self.head_size = _HEAD_SIZE
        self.page_size = _PAGE_SIZE
        self.idx_len = _IDX_LEN
        self.log_len = _LOG_LEN


    def load_config_from_head(self):
        """
        如果没有在初始化阶段对齐，则需要从数据库中读取对齐配置并对齐
        :return:
        """
        head_content = self.get_head()
        self.head_size = head_content['HEAD_SIZE']
        self.page_size = head_content['PAGE_SIZE']
        self.idx_len = head_content['IDX_LEN']
        self.log_len = head_content['LOG_LEN']
        return

    def __log_hook(self, hook_type=0, user=None, time=None, info=None): # 0:data 1:index 2:head(at build)
        self.append_log(log_content={'type':hook_type, 'user':user, 'time':time.time() if time is None else time, 'info':info})
        pass

    def get_head(self):
        head_content = _READ_HEAD(self.db, self.head_size)
        length = struct.unpack('<i', head_content[:4])[0]
        dic_content = head_content[4:4+length]
        return pickle.loads(dic_content)

    def compile_head(self, idx_len, log_len, head_size, page_size, version, info=None):
        self.set_head({'version':version,
                       'info':info,
                       'log_idx':0,
                       'idx_len':0,
                       'data_idx':0,
                       'LOG_LEN':log_len,
                       'IDX_LEN':idx_len,
                       'HEAD_SIZE':head_size,
                       'PAGE_SIZE':page_size})
        self.__log_hook(hook_type=2, time=time.time())
        return

    def set_head(self, dic):
        dic_bytes = pickle.dumps(dic)
        len_dic_bytes = len(dic_bytes)
        # print(len_dic_bytes)
        length = struct.pack('<i', len_dic_bytes)
        content = length + dic_bytes
        _WRITE_HEAD(_DB=self.db, _HEAD_CONTENT=content, _HEAD_SIZE=self.head_size)
        return

    def append_log(self, log_content):
        log_content = pickle.dumps(log_content)
        log_content_length = len(log_content)
        log_bytes = struct.pack('<i', log_content_length)+log_content
        head_info = self.get_head()
        log_idx = head_info['log_idx']
        _CONTENT_SET(_CONTENT=log_bytes,
                     _DB=self.db,
                     _PAGE_SIZE=self.page_size,
                     _TYPE=0x01,
                     _INDEX_LEN=self.idx_len,
                     _LOG_LEN=self.log_len,
                     _HEAD_SIZE=self.head_size,
                     _LOG_OFFSET=log_idx,)
        head_info['log_idx'] += len(log_bytes)
        self.set_head(head_info)
        return 0

    def check_log(self):
        log_lst = []
        head_info = self.get_head()
        log_max = head_info['log_idx']

        all_log_content = _CONTENT_MULTIPAGE(_DB=self.db,
                                     _CONTENT_LEN=log_max,
                                     _HEAD_SIZE=self.head_size,
                                     _PAGE_SIZE=self.page_size,
                                     _RET=0x01,
                                     _INDEX_LEN=self.idx_len,
                                     _LOG_LEN=self.log_len,
                                     _START_OFFSET=0)
        offset = 0
        while offset<len(all_log_content):
            log_len = struct.unpack('<i', all_log_content[offset:offset+4])[0]
            # print('log_len', log_len)
            log_content = pickle.loads(all_log_content[offset+4:offset+4+log_len])

            # print('log_len', log_content)
            # log_content = pickle.loads(log_content)
            log_lst.append(log_content)
            offset += 4+log_len
        return log_lst


    def update_index(self, idx_dic):
        idx_bytes = pickle.dumps(idx_dic)
        _CONTENT_SET(_CONTENT=idx_bytes,
                     _DB=self.db,
                     _PAGE_SIZE=self.page_size,
                     _TYPE=0x00,
                     _INDEX_LEN=self.idx_len,
                     _LOG_LEN=self.log_len,
                     _HEAD_SIZE=self.head_size,)
        head_info = self.get_head()
        head_info['idx_len'] = len(idx_bytes)
        self.set_head(head_info)
        self.__log_hook(hook_type=1, time=time.time())
        pass

    def get_index(self):
        head_info = self.get_head()
        idx_len = head_info['idx_len']
        if idx_len > 0:
            index_content = _CONTENT_MULTIPAGE(_DB=self.db,
                               _CONTENT_LEN=idx_len,
                               _HEAD_SIZE=self.head_size,
                               _PAGE_SIZE=self.page_size,
                               _RET=0x00,
                               _INDEX_LEN=self.idx_len,
                               _LOG_LEN=self.log_len,
                               _START_OFFSET=0)
            return pickle.loads(index_content)
        else:
            self.update_index({})
            return {}


    def get_data(self, offset):
        data_length = _CONTENT_MULTIPAGE(_DB=self.db,
                                           _CONTENT_LEN=4,
                                           _HEAD_SIZE=self.head_size,
                                           _PAGE_SIZE=self.page_size,
                                           _RET=0x02,
                                           _INDEX_LEN=self.idx_len,
                                           _LOG_LEN=self.log_len,
                                           _START_OFFSET=offset)
        data_length = pickle.unpack('<i', data_length)[0]
        # print(data_length)
        data_content = _CONTENT_MULTIPAGE(_DB=self.db,
                           _CONTENT_LEN=data_length,
                           _HEAD_SIZE=self.head_size,
                           _PAGE_SIZE=self.page_size,
                           _RET=0x02,
                           _INDEX_LEN=self.idx_len,
                           _LOG_LEN=self.log_len,
                           _START_OFFSET=offset+4)
        # print(len(data_content))
        return pickle.loads(data_content)

    def set_data(self, data, key=None):
        data_bytes = pickle.dumps(data)
        data_bytes_length = struct.pack('<i', len(data_bytes))
        head_info = self.get_head()
        data_idx = head_info['data_idx']
        data_content= data_bytes_length+data_bytes

        _CONTENT_SET(_CONTENT=data_content,
                     _DB=self.db,
                     _PAGE_SIZE=self.page_size,
                     _TYPE=0x02,
                     _INDEX_LEN=self.idx_len,
                     _LOG_LEN=self.log_len,
                     _HEAD_SIZE=self.head_size,
                     _DATA_OFFSET=data_idx)
        head_info['data_idx'] += len(data_content)
        self.set_head(head_info)
        self.__log_hook(hook_type=0, time=time.time(), info={key:hash(data_bytes)} if key is not None else None)
        return data_idx

    def get_current_position(self):
        return self.db.tell()

    @staticmethod
    def connect(db_name, head_size):
        db, status = _CONN_DATABASE(_NAME=db_name, _HEAD_SIZE=head_size)
        return db, status

    def close(self):
        self.db.close()


if __name__ == '__main__':

    # print(struct.unpack('<i', struct.pack('<i', 1))[0])
    db, status = Session.connect('test5.db', head_size=1000, )
    interface = Session(db, 1000, 50, 10, 10)
    # interface.set_head(dic={'version':'0.0.1beta', 'info':'qe', 'log_idx':0, 'idx_len':0, 'data_idx':0, 'LOG_LEN':10,'IDX_LEN':10,'HEAD_SIZE':1000, 'PAGE_SIZE':50})
    # print(interface.get_head()) # write-read test ok
    # interface.update_index({'a':'ggggggggg'}) # index 读写测试成功
    print(interface.get_index())
    # print(interface.get_head())
    # print(interface.get_head())
    # interface.set_data({'data1':23908450})
    # print(interface.set_data('hhhhhhhh23'))
    # print(interface.set_data('hh2gbvfopopopopopopophh23'))
    # print(interface.get_data(offset=0)) # data 读写调试成功
    # print(interface.get_data(offset=204))
    # print(interface.get_head())




import setuptools

with open("Odyssey/README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="OdysseyDB",  # Replace with your own username
    version="1.1.0",
    author="Yixiao Lan",
    author_email="yixiaolan@foxmail.com",
    description="A lightweight migratable file based key-value database with redis-like api and more features.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Eathoublu/OdysseyDB",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
)

